#include "precompile.h"

#include "building.h"
#include "collider.h"
#include "metamgr.h"
#include "room.h"
#include "loot.h"

Building::Building():Entity()
{
    entity_type = ET_Building;
}

Building::~Building()
{
}

void Building::Initialize()
{
    Entity::Initialize();
    RecalcSelfCollider();
}

void Building::RecalcSelfCollider()
{
    for (auto& obj : meta->i->staticobj()) {
        AabbCollider* collider = new AabbCollider();
        collider->owner = this;
        collider->_min = Vector2D(obj.x() - obj.width()/2.0 - meta->i->tilewidth()/2.0,
                                  obj.y() - obj.height()/2.0 - meta->i->tileheight()/2.0);
        collider->_max = Vector2D(obj.x() + obj.width()/2.0 - meta->i->tilewidth()/2.0,
                                  obj.y() + obj.height()/2.0 - meta->i->tileheight()/2.0);
        colliders.push_back(collider);
    }
}

void Building::FillMFObjectPart(cs::MFObjectPart* part_data)
{
    part_data->set_object_type(ET_Building);
    cs::MFBuildingPart* p = part_data->mutable_union_obj_3();
    p->set_obj_uniid(entity_uniid);
    pos.ToPB(p->mutable_pos());
}

void Building::FillMFObjectFull(cs::MFObjectFull* full_data)
{
    full_data->set_object_type(ET_Building);
    cs::MFBuildingFull* p = full_data->mutable_union_obj_3();
    p->set_obj_uniid(entity_uniid);
    pos.ToPB(p->mutable_pos());

    p->set_building_id(meta->i->mapid());
}

ColliderComponent* Building::GetBoxBound()
{
    AabbCollider* collider = new AabbCollider();
    collider->active = true;
    collider->owner = this;
    collider->_min = Vector2D(
                             -meta->i->tilewidth()/2.0,
                             -meta->i->tileheight()/2.0
                             );
    collider->_max = Vector2D(
                              meta->i->tilewidth()/2.0,
                              meta->i->tileheight()/2.0
                              );
    return collider;
}
