#include "precompile.h"

#include "entity.h"
#include "collider.h"
#include "collision.h"

bool ColliderComponent::IntersectSegment(Vector2D& p0, Vector2D& p1)
{
    switch (type) {
    case CT_None:
        break;
    case CT_Aabb:
        {
            AabbCollider* a_aabb = (AabbCollider*)this;
            return IntersectSegmentAabb(p0,
                                        p1,
                                        a_aabb->owner->pos + a_aabb->_min,
                                        a_aabb->owner->pos + a_aabb->_max);
        }
        break;
    case CT_Circle:
        {
            CircleCollider* a_circle = (CircleCollider*)this;
            return IntersectSegmentCircle(p0,
                                          p1,
                                          a_circle->owner->pos + a_circle->pos,
                                          a_circle->rad);
        }
        break;
    }
    return false;
}

bool ColliderComponent::Intersect(ColliderComponent* b)
{
    switch (type) {
    case CT_None:
        break;
    case CT_Aabb:
        {
            AabbCollider* a_aabb = (AabbCollider*)this;
            switch (b->type) {
            case CT_None:
                break;
            case CT_Aabb:
                {
                    AabbCollider* b_aabb = (AabbCollider*)b;
                    return IntersectAabbAabb(a_aabb->owner->pos + a_aabb->_min,
                                             a_aabb->owner->pos + a_aabb->_max,
                                             b_aabb->owner->pos + b_aabb->_min,
                                             b_aabb->owner->pos + b_aabb->_max
                                             );
                }
                break;
            case CT_Circle:
                {
                    CircleCollider* b_circle = (CircleCollider*)b;
                    return IntersectAabbCircle(a_aabb->owner->pos + a_aabb->_min,
                                               a_aabb->owner->pos + a_aabb->_max,
                                               b_circle->owner->pos + b_circle->pos,
                                               b_circle->rad);
                }
                break;
            }
        };
        break;
    case CT_Circle:
        {
            CircleCollider* a_circle = (CircleCollider*)this;
            switch (b->type) {
            case CT_None:
                break;
            case CT_Aabb:
                {
                    AabbCollider* b_aabb = (AabbCollider*)b;
                    return IntersectAabbCircle(b_aabb->owner->pos + b_aabb->_min,
                                               b_aabb->owner->pos + b_aabb->_max,
                                               a_circle->owner->pos + a_circle->pos,
                                               a_circle->rad);
                }
                break;
            case CT_Circle:
                {
                    CircleCollider* b_circle = (CircleCollider*)b;
                    return IntersectCircleCircle(
                                                 a_circle->owner->pos + a_circle->pos,
                                                 a_circle->rad,
                                                 b_circle->owner->pos + b_circle->pos,
                                                 b_circle->rad);
                }
                break;
            }
        }
        break;
    }
    return false;
}

bool ColliderComponent::CalcSafePoint(ColliderComponent* b, Vector2D& new_pos)
{
    switch (type) {
    case CT_None:
        break;
    case CT_Aabb:
        {
            AabbCollider* a_aabb = (AabbCollider*)this;
            switch (b->type) {
            case CT_Aabb:
                {
                    AabbCollider* b_aabb = (AabbCollider*)this;
                    return CalcAabbAabbSafePoint(a_aabb->owner->pos + a_aabb->_min,
                                                 a_aabb->owner->pos + a_aabb->_max,
                                                 b_aabb->owner->pos + b_aabb->_min,
                                                 b_aabb->owner->pos + b_aabb->_max,
                                                 new_pos);
                }
                break;
            case CT_Circle:
                {
                    CircleCollider* b_circle = (CircleCollider*)b;
                    return CalcAabbAabbSafePoint(a_aabb->owner->pos + a_aabb->_min,
                                                 a_aabb->owner->pos + a_aabb->_max,
                                                 b_circle->owner->pos + b_circle->pos,
                                                 b_circle->rad,
                                                 new_pos);
                }
                break;
            }
        }
        break;
    case CT_Circle:
        {
            CircleCollider* a_circle = (CircleCollider*)this;
            switch (b->type) {
            case CT_Aabb:
                {
                    AabbCollider* b_aabb = (AabbCollider*)b;
                    return CalcCircleAabbSafePoint(
                                                   a_circle->owner->pos + a_circle->pos,
                                                   a_circle->rad,
                                                   b_aabb->owner->pos + b_aabb->_min,
                                                   b_aabb->owner->pos + b_aabb->_max,
                                                   new_pos);
                }
                break;
            case CT_Circle:
                {
                    CircleCollider* b_circle = (CircleCollider*)b;
                    return CalcCircleCircleSafePoint(
                                                     a_circle->owner->pos + a_circle->pos,
                                                     a_circle->rad,
                                                     b_circle->owner->pos + b_circle->pos,
                                                     b_circle->rad,
                                                     new_pos);
                }
                break;
            }
        }
        break;
    default:
        break;
    }
    return false;
}

void DestoryCollider(ColliderComponent* collider)
{
    switch (collider->type) {
    case CT_Aabb:
        {
            delete (AabbCollider*)collider;
        }
        break;
    case CT_Circle:
        {
            delete (CircleCollider*)collider;
        }
        break;
    default:
        {
            delete collider;
        }
        break;
    }
}
